//
// Copyright 2001,2002 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#ifndef _Joystick_h_DEFINED
#define _Joystick_h_DEFINED

#include <OPENR/OObject.h>
#include <OPENR/OSubject.h>
#include <OPENR/OObserver.h>
#include <OPENR/ODataFormats.h>

#include <OMWares/OMoNetMessage.h>
#include <OMWares/MoNetMessageMaker.h>
#include <OMWares/MMM_CommandList.h>

#include <MoNetData.h>

#include "def.h"
#include "MNetMsgQ.h"

// The possible states
enum JoystickBehavior {
    J_BOOTING, // Booting up
    J_ONLINE,  // Normal mode, read commands from joystick/keyboard
    J_PREP_SM, // Preparing special motion
    J_EXEC_SM, // Executing special motion
    J_RECOVER  // Recovering from fall
};

// OMoNet command = M_*
// MoNet command = SM_*  (special motion)
// MoNet commands found at MS/OPEN-R/MW/CONF/MONETCMD.CFG
enum JoystickMotion {
  M_UNDEFINED,      // 0
  M_RECOVERFALL,    // 1
  M_STOP_E,         // 2
  M_STOP_N,         // 3
  M_FORWARD,        // 4
  M_RIGHTFORWARD,   // 5
  M_LEFTFORWARD,    // 6
  M_RIGHT,          // 7
  M_LEFT,           // 8
  M_BACKWARD,       // 9
  M_TURNRIGHT,      // 10
  M_TURNLEFT,       // 11
  M_RIGHTKICK,      // 12
  M_LEFTKICK,       // 13
  M_STAND,          // 14
  M_RESET,          // 15
  SM_SLEEP,         // 16
  SM_SIT,           // 17
  SM_STAND,         // 18
  SM_CHEER,         // 19
  SM_DANCE1,        // 20
  SM_DANCE2,        // 21
  SM_DANCE3,        // 22
  SM_SITBYE,        // 23
  SM_SITCOME,       // 24
  SM_SITDANCE,      // 25
  SM_SITLOOSE,      // 26
  SM_SITPURR,       // 27
  SM_SITWIN,        // 28
  SM_SLEEPDENY,     // 29
  SM_STANDBYE,      // 30
  SM_WAV_WIN,       // 31
  SM_WAV_LOOSE,     // 32
  SM_WAV_MTN_BYE1,  // 33
  SM_WAV_MTN_BYE2,  // 34
  SM_WAV_MTN_SLEEP  // 35
};

// MoNet commands is number 16 to 35 in JoystickMotion (used in: Joystick::MakeDecision)
#define SMNUM_FROM 16
#define SMNUM_TO 35

// Modify these definitions to change the connection: ButtonNr <-> Action
enum Button {
    B_SPECIALACTION, // Button 1
    B_RESET,         // --<>-- 2
    B_LEFTKICK,      // --<>-- 3
    B_RIGHTKICK,     // --<>-- 4
    B_BUTTON5,       // --<>-- 5
    B_TAILSWING,     // --<>-- 6
    B_DISENGAGE,     // --<>-- 7
    B_QUIT           // --<>-- 8
};

// Modify these definitions to change the connection: Axisdirection <-> Action
enum Axis {
    A_LEFTRIGHT,  // Axis 1
    A_UPDOWN,     // -<>- 2
    A_TURN,       // -<>- 3
    A_CYLINDER,   // -<>- 4
    A_2LEFTRIGHT, // -<>- 5
    A_2UPDOWN,    // -<>- 6
};

// The axis movement required for the movement to be notyfied
#define J_THRESHOLD 1

// The speed of the headmovement
#define PAN_GAIN  2
#define TILT_GAIN 2

class Joystick : public OObject {
public:

    Joystick();
    virtual ~Joystick() {}

    OSubject*  subject[numOfSubject];
    OObserver* observer[numOfObserver];
        
    virtual OStatus DoInit    (const OSystemEvent& event);
    virtual OStatus DoStart   (const OSystemEvent& event);
    virtual OStatus DoStop    (const OSystemEvent& event);
    virtual OStatus DoDestroy (const OSystemEvent& event);
    
    void ReadyMotion         (const OReadyEvent& event);
    void ReadySpecialMotion  (const OReadyEvent& event);
    void NotifySpecialMotion (const ONotifyEvent& event);
    void NotifyMotion        (const ONotifyEvent& event);
    void ResultGSensor       (const ONotifyEvent& event);
    void ResultSensor        (const ONotifyEvent& event);
    void MakeDecision        (const ONotifyEvent& event);
    

private:
    JoystickBehavior  jbehavior_;
    JoystickMotion    jmotion_;
  
    MoNetMessageMaker mmm_;     // To make OMoNetMessage
    MNetMsgQ          msgQ_;    // OMoNetMessage-queue
    
    int headtilt_;             // Tilt angle of head (degrees)
    int headpan_;              // Pan angle of head (degrees)
    bool smTryAgain_;
    bool moveHeadUp;
    bool moveHeadDown;
    bool moveHeadLeft;
    bool moveHeadRight;

    OStatus AddToMessageQueue();
    OStatus AddToMessageQueue(OMoNetMessage message);
    void SendMotion();
    void Move(); // Moves according to 'jmotion_'-variable
    void MoveHead(int tilt, int pan, int roll);
    void MoveHead(int tilt, int pan);
    void SwingTail(int tilt, int pan);
    void SwingTail();
    void StopTail();
    int GetCommand(); // Returns MoNet command ID according to current jmotion_ value
    void GetUp();
    void SendStandMotion();
    void SendResetMotion();
    void PrepareSpecialMotion();
    void ExecSpecialMotion();
};

#endif // _Joystick_h_DEFINED
