//
// Copyright 2002 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/core_macro.h>
#include <OPENR/OFbkImage.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include "Server.h"
#include "entry.h"

byte mybuffer[20000];

Server::Server():readySend(true), fbkID(oprimitiveID_UNDEF)
{
}

OStatus
Server::DoInit(const OSystemEvent& event)
{
    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;
    OpenImagePrimitive();
    return oSUCCESS;
}

OStatus
Server::DoStart(const OSystemEvent& event)
{ 
    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;
        
    ipstackRef = antStackRef("IPStack");
    pictureSend = -1;

    for (int index = 0; index < SERVER_CONNECTION_MAX; index++) {
        OStatus result = InitTCPConnection(index);
        Listen(index);
	if (result != oSUCCESS) return oFAIL;
    }

    //Listen(0);

    return oSUCCESS;
}    

OStatus
Server::DoStop(const OSystemEvent& event)
{
    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;
    return oSUCCESS;
}

OStatus
Server::DoDestroy(const OSystemEvent& event)
{
    DELETE_ALL_SUBJECT_AND_OBSERVER;
    return oSUCCESS;
}

void
Server::OpenImagePrimitive()
{
  pictureSend = -1;
    OStatus result = OPENR::OpenPrimitive(FBK_LOCATOR, &fbkID);
    if (result != oSUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "ImageObserver::OpenPrimitive()",
                  "OPENR::OpenPrimitive() FAILED", result));
    }
}

void
Server::Ready(const OReadyEvent& event)
{
  readySend = true;
}

OStatus
Server::Listen(int index)
{
     if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    //
    // Create endpoint
    //
    antEnvCreateEndpointMsg tcpCreateMsg(EndpointType_TCP,
                                         SERVER_BUFFER_SIZE * 2);
    tcpCreateMsg.Call(ipstackRef, sizeof(tcpCreateMsg));
    if (tcpCreateMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Server::Listen()",
                  "Can't create endpoint",
                  index, tcpCreateMsg.error));
        return oFAIL;
    }
    connection[index].endpoint = tcpCreateMsg.moduleRef;

    //
    // Listen
    //
    TCPEndpointListenMsg listenMsg(connection[index].endpoint,
                                   IP_ADDR_ANY, SERVER_PORT);
    listenMsg.continuation = (void*)index;

    listenMsg.Send(ipstackRef, myOID_,
                   Extra_Entry[entryListenCont], sizeof(listenMsg));
    
    connection[index].state = CONNECTION_LISTENING;
    return oSUCCESS;
}

void
Server::ListenCont(ANTENVMSG msg)
{
    TCPEndpointListenMsg* listenMsg = (TCPEndpointListenMsg*)antEnvMsg::Receive(msg);
    int index = (int)listenMsg->continuation;

    if (listenMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "Server::ListenCont()",
                  "FAILED. listenMsg->error", listenMsg->error));
        Close(index);
        return;
    }

    connection[index].state = CONNECTION_CONNECTED;

    SetReceiveData(index);
    Receive(index);
}

OStatus
Server::Send(int index)
{

    if (connection[index].sendSize == 0 ||
        connection[index].state != CONNECTION_CONNECTED) return oFAIL;

    TCPEndpointSendMsg sendMsg(connection[index].endpoint,
                               connection[index].sendData,
                               connection[index].sendSize);
    sendMsg.continuation = (void*)index;

    sendMsg.Send(ipstackRef, myOID_,
                 Extra_Entry[entrySendCont],
                 sizeof(TCPEndpointSendMsg));

    connection[index].state = CONNECTION_SENDING;
    connection[index].sendSize = 0;

    return oSUCCESS;
}

void
Server::SendCont(ANTENVMSG msg)
{
  TCPEndpointSendMsg* sendMsg = (TCPEndpointSendMsg*)antEnvMsg::Receive(msg);
  int index = (int)(sendMsg->continuation);
  
  if (sendMsg->error != TCP_SUCCESS) {
    OSYSLOG1((osyslogERROR, "%s : %s %d",
	      "Server::SendCont()",
	      "FAILED. sendMsg->error", sendMsg->error));
    Close(index);
    return;
  }

  connection[index].state = CONNECTION_CONNECTED;

  if (pictureSend == index) { 
    if (total_sent < 19008) {
      byte temp_buf[SIZE_OF_PACKAGE];
      for (int i = 0; i < SIZE_OF_PACKAGE; i++)
	temp_buf[i] = mybuffer[i+total_sent];
      memcpy(connection[index].sendData,&temp_buf,SIZE_OF_PACKAGE);
      connection[index].sendSize = SIZE_OF_PACKAGE;
      total_sent += SIZE_OF_PACKAGE;
      Send(index);
      return;
    }
    else 
      pictureSend = -1;
  }
  
  //OSYSPRINT(("sendData : %s", connection[index].sendData));
 
  SetReceiveData(index);
  Receive(index);
}

OStatus
Server::Receive(int index)
{
    if (connection[index].state != CONNECTION_CONNECTED &&
        connection[index].state != CONNECTION_SENDING) return oFAIL;

    TCPEndpointReceiveMsg receiveMsg(connection[index].endpoint,
                                     connection[index].recvData,
                                     1, connection[index].recvSize);
    receiveMsg.continuation = (void*)index;

    receiveMsg.Send(ipstackRef, myOID_,
                    Extra_Entry[entryReceiveCont], sizeof(receiveMsg));

    return oSUCCESS;
}


void 
Server::PictureNotify(const ONotifyEvent& event) 
{
  OSYSPRINT(("PictureSend: %d\n",pictureSend));
  if (pictureSend != -1) {
    // Send picture
    OFbkImageVectorData* pictureData  = (OFbkImageVectorData*)event.Data(0);
    int index = pictureSend;
    byte* pd = pictureData->GetData(ofbkimageLAYER_M);
    OFbkImageInfo* info = pictureData->GetInfo(ofbkimageLAYER_M);
    size_t w = info->width;
    size_t h = info->height;
    size_t s = 3 * info->width * info->height;
    //byte mybuffer[20000];
    if (pd != NULL) {
      int myindex = 0;
      OFbkImage yImage(info, pd, ofbkimageBAND_Y);
      OFbkImage uImage(info, pd, ofbkimageBAND_Cr);
      OFbkImage vImage(info, pd, ofbkimageBAND_Cb);
      
      for (int y = 0; y < yImage.Height(); y++) {
	for (int x = 0; x < yImage.Width(); x++) {
	  mybuffer[myindex++]=yImage.Pixel(x,y);
	  mybuffer[myindex++]=uImage.Pixel(x,y);
	  mybuffer[myindex++]=vImage.Pixel(x,y);
	}
      }

      byte temp_buf[SIZE_OF_PACKAGE];
      for (int i = 0; i < SIZE_OF_PACKAGE; i++)
	temp_buf[i] = mybuffer[i];
      memcpy(connection[index].sendData,&temp_buf,SIZE_OF_PACKAGE);
      connection[index].sendSize = SIZE_OF_PACKAGE;
      total_sent = SIZE_OF_PACKAGE;
      Send(index);
      
      //for(int i = 0; i < 100000; i++) {}
        
      //memcpy(connection[index].sendData,&mybuffer,s);
      //for (int i = 0; i < 100000; i++) {}
      /*
      int c = 10000;
      while (1) {
	OSYSPRINT(("%d: %d\n",c,connection[index].sendData[c]));
	if (c > 10020)
	break;
	c++;
      }
      */
      connection[index].sendSize = s;
      Send(index);
    }
  }
  //observer[event.ObsIndex()]->AssertReady(event.SenderID());
}

void
Server::ReceiveCont(ANTENVMSG msg)
{
  //OSYSPRINT(("ReciveCont\n"));
  TCPEndpointReceiveMsg* receiveMsg = (TCPEndpointReceiveMsg*)antEnvMsg::Receive(msg);
    int index = (int)(receiveMsg->continuation);

    if (receiveMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "EchoServer::ReceiveCont()",
                  "FAILED. receiveMsg->error", receiveMsg->error));
        Close(index);
        return;
    }
    
    //OSYSPRINT(("Server has recived: %s\n", connection[index].recvData));
    char command[256];
    sprintf(command,"%s",connection[index].recvData);	

    if (!strncmp(command,"X",1)) {
      // Take a picture
      OSYSPRINT(("Take a picture\n"));
      pictureSend = index;
      observer[obsPicture]->AssertReady();
      return;
    }

    if (readySend) {
      readySend = false;
      subject[sbjSendString]->SetData(command, sizeof(command));
      subject[sbjSendString]->NotifyObservers();
    } 
    else
      OSYSPRINT(("Not sent to joystick, it's busy!\n"));

    connection[index].sendSize = receiveMsg->sizeMin;
    //OSYSPRINT(("connection[index].sendSize: %d\n",connection[index].sendSize));
    SetSendData(index);
    Send(index);
    OSYSPRINT(("Send something\n"));
}

OStatus
Server::Close(int index)
{
    if (connection[index].state == CONNECTION_CLOSED ||
        connection[index].state == CONNECTION_CLOSING) return oFAIL;

    TCPEndpointCloseMsg closeMsg(connection[index].endpoint);
    closeMsg.continuation = (void*)index;

    closeMsg.Send(ipstackRef, myOID_,
                  Extra_Entry[entryCloseCont], sizeof(closeMsg));

    connection[index].state = CONNECTION_CLOSING;

    return oSUCCESS;
}

void
Server::CloseCont(ANTENVMSG msg)
{
    TCPEndpointCloseMsg* closeMsg = (TCPEndpointCloseMsg*)antEnvMsg::Receive(msg);
    int index = (int)(closeMsg->continuation);

    connection[index].state = CONNECTION_CLOSED;
    Listen(index);
}

OStatus
Server::InitTCPConnection(int index)
{
    connection[index].state = CONNECTION_CLOSED;

    // 
    // Allocate send buffer
    //
    antEnvCreateSharedBufferMsg sendBufferMsg(SERVER_BUFFER_SIZE);

    sendBufferMsg.Call(ipstackRef, sizeof(sendBufferMsg));
    if (sendBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Server::InitTCPConnection()",
                  "Can't allocate send buffer",
                  index, sendBufferMsg.error));
        return oFAIL;
    }

    connection[index].sendBuffer = sendBufferMsg.buffer;
    connection[index].sendBuffer.Map();
    connection[index].sendData
        = (byte*)(connection[index].sendBuffer.GetAddress());

    //
    // Allocate receive buffer
    //
    antEnvCreateSharedBufferMsg recvBufferMsg(SERVER_BUFFER_SIZE);

    recvBufferMsg.Call(ipstackRef, sizeof(recvBufferMsg));
    if (recvBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Server::InitTCPConnection()",
                  "Can't allocate receive buffer",
                  index, recvBufferMsg.error));
        return oFAIL;
    }

    connection[index].recvBuffer = recvBufferMsg.buffer;
    connection[index].recvBuffer.Map();
    connection[index].recvData
        = (byte*)(connection[index].recvBuffer.GetAddress());

    return oSUCCESS;
}

void
Server::SetSendData(int index)
{
    memcpy(connection[index].sendData,
           connection[index].recvData, connection[index].sendSize);
}

void
Server::SetReceiveData(int index)
{
    connection[index].recvSize = SERVER_BUFFER_SIZE;
}
